#ifndef VECTOR_H_
#define VECTOR_H_

#include <cassert>
#include <cmath>
#include <string>
#include <iostream>
#include <sstream>


namespace arda 
   {
   ////////////////////////////////////////////////////////////////////////////////
   // Vectors
   //
   // Templated classes for 2, 3, and 4 dimensional vectors of any type,
   //     although these types are really intended to be int, float, and double 
   //     only.  (i.e. if you roll your own base type, it had better act like
   //     a numeric scalar.)
   //
   // Supported operations on vectors
   //
   // Construction
   //     Vector v;             // Default zero vector.
   //     Vector v2(v);         // Construct from other vector.
   //     T a, b;
   //     Vector v(a, b)        // Construct from values.
   // []
   //     If you have a Vector named v you can access it's member elements as
   //     v[0], v[1], v[2], and v[3]. in addtion to v.x, v.y, v.z, and v.w
   // == != + += - -=
   //     Defined for operations on two vectors of the same type.
   // * *= / /=
   //     Defined for scalar multiplication/division with int, float, and double.
   //     * is defined as a standalone template operator for the scalar * Vector form.
   // assign(T a, T b [, T b [, T d]])
   //     Directly assign values to the vector.
   // cross(Vector3& v2)
   // cross(Vector3& v2, Vector3& vres)
   //     Cross product of the vector with v2.  Only defined for Vector3 types.
   //     The version that takes two arguments stores the result in the second
   //     argument.  This may be faster since no temporary object is created
   //     during the operation.
   // dot(Vector& v2)
   //     The dot product of the vector with v2.
   // get_angle(Vector& v2)
   //     Returns the angle (in radians) between the vector and v2.
   //     Not meaningful for int vectors.
   // get_anglen(Vector& v2)
   //     Like get_angle(), but the vectors must already be normalized.
   //     Not meaningful for int vectors.
   // getstring(std::string & s)
   //     Returns a C string representation of the vector.
   //     This is one of the few non-inline Vector functions.
   // length()
   //     Get the length of a vector.
   //     Returns a double for all vector types.
   // normalize()
   //     Normalizes the vecor.
   //     Not meaningful for int vectors.
   // proj(Vector& v2, Vector& vres)
   //     Calculates the projection of the vector onto v2 and returns the result 
   //     in vres. 
   //     Not meaningful for int vectors.
   //     Returns vres.
   //
   //
   // Note that the following typedefs are defined for convenience:
   //
   // typedef Vector2<int> Vector2i;
   // typedef Vector2<float> Vector2f;
   // typedef Vector2<double> Vector2d;
   // typedef Vector3<int> Vector3i;
   // typedef Vector3<float> Vector3f;
   // typedef Vector3<double> Vector3d;
   // typedef Vector4<int> Vector4i;
   // typedef Vector4<float> Vector4f;
   // typedef Vector4<double> Vector4d;
   //
   //
   //
   // Scalar multiplication and division are overloaded.  Why?
   // I'm not sure I need this.  The idea is to support all likely scalar
   // types without having to rely on implicit or explicit type conversion,
   // especially not a conversion that might result in loss of precision.
   // Multiplying a float Vector by a double scalar, for example.
   //
   // Is there a better way to do this?  Am I worrying about nothing?
   //
   //
   //
   // I suppose I could make this into a single Vector template on two parameters
   // template <typename T, unsigned int N> class Vector
   //    {
   //    T v[N];
   //
   //    etc...
   //
   // This might buy perfect generality, but it would come at the expense of
   // loops in all of the code, so I don't think it would really be worth it.


   /////////////////////////////////////////////////////////////////////////////
   template <typename T> class Vector2
      {
      public:
      T x, y;

      // Constructors
      Vector2() {}
      Vector2(T a) : x (a), y(a) {}
      Vector2(T a, T b) : x (a), y(b) {}

      // Array indexing
      inline T& operator[](unsigned int i)
	 { assert (i<2); if (i==0) return x; return y; }
      inline const T& operator[](unsigned int i) const
	 { assert (i<2); if (i==0) return x; return y; }

      // Assignment
      inline Vector2<T>& operator=(const Vector2<T>& v2)
	 { x = v2.x; y = v2.y; return *this; }
      inline Vector2<T>& assign(const T a, const T b)
	 { x = a; y = b; return *this; }

      // Comparison
      inline bool operator==(Vector2<T>& v2) const
	 { return ((x == v2.x) && (y == v2.y)); }
      inline bool operator!=(Vector2<T>& v2) const
	 { return ! (*this == v2); }

      // Vector addition
      inline Vector2<T>& operator+=(const Vector2<T>& v2)
	 { x += v2.x; y += v2.y; return *this; }
      inline Vector2<T> operator+(const Vector2<T>& v2) const
	 { return Vector2<T>(*this) += v2; }

      // Vector subtraction
      inline Vector2<T>& operator-=(const Vector2<T>& v2)
	 { x -= v2.x; y -= v2.y; return *this; }
      inline Vector2<T> operator-(const Vector2<T>& v2) const
	 { return Vector2<T>(*this) -= v2; }

      // Scalar multiplication
      inline Vector2<T>& operator*=(const int a)
	 { x *= a; y *= a; return *this; }
      inline Vector2<T>& operator*=(const float a)
	 { x *= a; y *= a; return *this; }
      inline Vector2<T>& operator*=(const double a)
	 { x *= a; y *= a; return *this; }
      inline Vector2<T> operator*(const int a)
	 { return Vector2<T>(*this) *= a;}
      inline Vector2<T> operator*(const float a)
	 { return Vector2<T>(*this) *= a;}
      inline Vector2<T> operator*(const double a)
	 { return Vector2<T>(*this) *= a;}

      // Scalar division
      inline Vector2<T>& operator/=(const int a)
	 { assert(a!=0); x /= a; y /= a; return *this; }
      inline Vector2<T>& operator/=(const float a)
	 { assert(a!=0); x /= a; y /= a; return *this; }
      inline Vector2<T>& operator/=(const double a)
	 { assert(a!=0); x /= a; y /= a; return *this; }
      inline Vector2<T> operator/(const int a)
	 { return Vector2<T>(*this) /= a;}
      inline Vector2<T> operator/(const float a)
	 { return Vector2<T>(*this) /= a;}
      inline Vector2<T> operator/(const double a)
	 { return Vector2<T>(*this) /= a;}


      // methods

      inline T dot(const Vector2<T>& v2) const
	 { return x*v2.x + y*v2.y; }

      inline double get_angle(Vector2<T>& v2)
	 { 
	 double tmp = dot(v2); 
	 return acos(sqrt(tmp*tmp/(dot(*this)*v2.dot(v2))));
	 }

      inline double get_anglen(Vector2<T>& v2)
	 { return acos((double) dot(v2)); }

      std::string & getstring(std::string & s) const;

      inline double length(void) const
	 { return sqrt((dot(*this))); }

      inline Vector2<T>& normalize()
	 { double l = length(); if (l == 0.0) return *this; x /= l; y /= l; return *this; }

      inline Vector2<T> proj(Vector2<T>& v2)
	 { Vector2<T> vres; vres = v2 * dot(v2)/v2.dot(v2); return vres; }
      inline Vector2<T>& proj(Vector2<T>& v2, Vector2<T>& vres)
	 { vres = v2 * dot(v2)/v2.dot(v2); return vres; }
      };

   // Scalar multiplication, continued
   template <typename T> inline Vector2<T> operator*(const int a, const Vector2<T>& v)
      { return Vector2<T>(v) *= a;}
   template <typename T> inline Vector2<T> operator*(const float a, const Vector2<T> v)
      { return Vector2<T>(v) *= a;}
   template <typename T> inline Vector2<T> operator*(const double a, const Vector2<T>& v)
      { return Vector2<T>(v) *= a;}


   /////////////////////////////////////////////////////////////////////////////
   template <typename T> class Vector3
      {
      public:
      T x, y, z;

      // Constructors
      Vector3() {}
      Vector3(T a) : x (a), y(a), z(a) {}
      Vector3(T a, T b, T c) : x (a), y(b), z(c) {}

      // Array indexing
      inline T& operator[](unsigned int i)
	 { assert (i<3); if (i==0) return x; if (i==1) return y; return z; }
      inline const T& operator[](unsigned int i) const
	 { assert (i<3); if (i==0) return x; if (i==1) return y; return z; }

      // Assignment
      inline Vector3<T>& operator=(const Vector3<T>& v2)
	 { x = v2.x; y = v2.y; z = v2.z; return *this; }
      inline Vector3<T>& assign(const T a, const T b, const T c)
	 { x = a; y = b; z = c; return *this; }

      // Comparison
      inline bool operator==(Vector3<T>& v2) const
	 { return ((x == v2.x) && (y == v2.y) && (z == v2.z)); }
      inline bool operator!=(Vector3<T>& v2) const
	 { return ! (*this == v2); }

      // Vector addition
      inline Vector3<T>& operator+=(const Vector3<T>& v2)
	 { x += v2.x; y += v2.y; z += v2.z; return *this; }
      inline Vector3<T> operator+(const Vector3<T>& v2) const
	 { return Vector3<T>(*this) += v2; }

      // Vector subtraction
      inline Vector3<T>& operator-=(const Vector3<T>& v2)
	 { x -= v2.x; y -= v2.y; z -= v2.z; return *this; }
      inline Vector3<T> operator-(const Vector3<T>& v2) const
	 { return Vector3<T>(*this) -= v2; }

      // Scalar multiplication
      inline Vector3<T>& operator*=(const int a)
	 { x *= a; y *= a; z *= a; return *this; }
      inline Vector3<T>& operator*=(const float a)
	 { x *= a; y *= a; z *= a; return *this; }
      inline Vector3<T>& operator*=(const double a)
	 { x *= a; y *= a; z *= a; return *this; }
      inline Vector3<T> operator*(const int a)
	 { return Vector3<T>(*this) *= a;}
      inline Vector3<T> operator*(const float a)
	 { return Vector3<T>(*this) *= a;}
      inline Vector3<T> operator*(const double a)
	 { return Vector3<T>(*this) *= a;}

      // Scalar division
      inline Vector3<T>& operator/=(const int a)
	 { assert(a!=0); x /= a; y /= a; z /= a; return *this; }
      inline Vector3<T>& operator/=(const float a)
	 { assert(a!=0); x /= a; y /= a; z /= a; return *this; }
      inline Vector3<T>& operator/=(const double a)
	 { assert(a!=0); x /= a; y /= a; z /= a; return *this; }
      inline Vector3<T> operator/(const int a)
	 { return Vector3<T>(*this) /= a;}
      inline Vector3<T> operator/(const float a)
	 { return Vector3<T>(*this) /= a;}
      inline Vector3<T> operator/(const double a)
	 { return Vector3<T>(*this) /= a;}


      // methods

      inline Vector3<T>& cross(Vector3<T>& v2, Vector3<T>& vres)
	 { 
	 vres.x =  y*v2.z - v2.y*z;
	 vres.y = -x*v2.z + v2.x*z;
	 vres.z =  x*v2.y - v2.x*y;
	 return vres;
	 }
      inline Vector3<T> cross(Vector3<T>& v2)
	 { 
	 Vector3<T> vres;
	 vres.x =  y*v2.z - v2.y*z;
	 vres.y = -x*v2.z + v2.x*z;
	 vres.z =  x*v2.y - v2.x*y;
	 return vres;
	 }

      inline T dot(const Vector3<T>& v2) const
	 { return x*v2.x + y*v2.y + z*v2.z; }

      inline double get_angle(Vector3<T>& v2)
	 { 
	 double tmp = dot(v2); 
	 return acos(sqrt(tmp*tmp/(dot(*this)*v2.dot(v2))));
	 }

      inline double get_anglen(Vector3<T>& v2)
	 { return acos((double) dot(v2)); }

      std::string & getstring(std::string & s) const;

      inline double length(void) const
	 { return sqrt((dot(*this))); }

      inline Vector3<T>& normalize()
	 { double l = length(); if (l == 0.0) return *this; x /= l; y /= l; z /= l; return *this; }

      inline Vector3<T> proj(Vector3<T>& v2)
	 { Vector3<T> vres; vres = v2 * dot(v2)/v2.dot(v2); return vres; }
      inline Vector3<T>& proj(Vector3<T>& v2, Vector3<T>& vres)
	 { vres = v2 * dot(v2)/v2.dot(v2); return vres; }
      };

   // Scalar multiplication, continued
   template <typename T> inline Vector3<T> operator*(const int a, const Vector3<T>& v)
      { return Vector3<T>(v) *= a;}
   template <typename T> inline Vector3<T> operator*(const float a, const Vector3<T> v)
      { return Vector3<T>(v) *= a;}
   template <typename T> inline Vector3<T> operator*(const double a, const Vector3<T>& v)
      { return Vector3<T>(v) *= a;}

   /////////////////////////////////////////////////////////////////////////////
   template <typename T> class Vector4
      {
      public:
      T x, y, z, w;

      // Constructors
      Vector4() {}
      Vector4(T a) : x (a), y(a), z(a), w(a) {}
      Vector4(T a, T b, T c, T d) : x (a), y(b), z(c), w(d) {}

      // Array indexing
      inline T& operator[](unsigned int i)
	 { assert (i<4); if (i==0) return x; if (i==1) return y; if (i==2) return z; return w; }
      inline const T& operator[](unsigned int i) const
	 { assert (i<4); if (i==0) return x; if (i==1) return y; if (i==2) return z; return w; }

      // Assignment
      inline Vector4<T>& operator=(const Vector4<T>& v2)
	 { x = v2.x; y = v2.y; z = v2.z; w = v2.w; return *this; }
      inline Vector4<T>& assign(const T a, const T b, const T c, const T d)
	 { x = a; y = b; z = c; w = d; return *this; }

      // Comparison
      inline bool operator==(Vector4<T>& v2) const
	 { return ((x == v2.x) && (y == v2.y) && (z == v2.z) && (w == v2.w)); }
      inline bool operator!=(Vector4<T>& v2) const
	 { return ! (*this == v2); }

      // Vector addition
      inline Vector4<T>& operator+=(const Vector4<T>& v2)
	 { x += v2.x; y += v2.y; z += v2.z; w += v2.w; return *this; }
      inline Vector4<T> operator+(const Vector4<T>& v2) const
	 { return Vector4<T>(*this) += v2; }

      // Vector subtraction
      inline Vector4<T>& operator-=(const Vector4<T>& v2)
	 { x -= v2.x; y -= v2.y; z -= v2.z; w -= v2.w; return *this; }
      inline Vector4<T> operator-(const Vector4<T>& v2) const
	 { return Vector4<T>(*this) -= v2; }

      // Scalar multiplication
      inline Vector4<T>& operator*=(const int a)
	 { x *= a; y *= a; z *= a; w *= a; return *this; }
      inline Vector4<T>& operator*=(const float a)
	 { x *= a; y *= a; z *= a; w *= a; return *this; }
      inline Vector4<T>& operator*=(const double a)
	 { x *= a; y *= a; z *= a; w *= a; return *this; }
      inline Vector4<T> operator*(const int a)
	 { return Vector4<T>(*this) *= a;}
      inline Vector4<T> operator*(const float a)
	 { return Vector4<T>(*this) *= a;}
      inline Vector4<T> operator*(const double a)
	 { return Vector4<T>(*this) *= a;}

      // Scalar division
      inline Vector4<T>& operator/=(const int a)
	 { assert(a!=0); x /= a; y /= a; z /= a; w /= a; return *this; }
      inline Vector4<T>& operator/=(const float a)
	 { assert(a!=0); x /= a; y /= a; z /= a; w /= a; return *this; }
      inline Vector4<T>& operator/=(const double a)
	 { assert(a!=0); x /= a; y /= a; z /= a; w /= a; return *this; }
      inline Vector4<T> operator/(const int a)
	 { return Vector4<T>(*this) /= a;}
      inline Vector4<T> operator/(const float a)
	 { return Vector4<T>(*this) /= a;}
      inline Vector4<T> operator/(const double a)
	 { return Vector4<T>(*this) /= a;}


      // methods

      inline T dot(const Vector4<T>& v2) const
	 { return x*v2.x + y*v2.y + z*v2.z + w*v2.w; }

      inline double get_angle(Vector4<T>& v2)
	 { 
	 double tmp = dot(v2); 
	 return acos(sqrt(tmp*tmp/(dot(*this)*v2.dot(v2))));
	 }

      inline double get_anglen(Vector4<T>& v2)
	 { return acos((double) dot(v2)); }

      std::string & getstring(std::string & s) const;

      inline double length(void) const
	 { return sqrt((dot(*this))); }

      inline Vector4<T>& normalize()
	 { double l = length(); if (l == 0.0) return *this; x /= l; y /= l; z /= l; w /= l; return *this; }

      inline Vector4<T> proj(Vector4<T>& v2)
	 { Vector4<T> vres; vres = v2 * dot(v2)/v2.dot(v2); return vres; }
      inline Vector4<T>& proj(Vector4<T>& v2, Vector4<T>& vres)
	 { vres = v2 * dot(v2)/v2.dot(v2); return vres; }
      };

   // Scalar multiplication, continued
   template <typename T> inline Vector4<T> operator*(const int a, const Vector4<T>& v)
      { return Vector4<T>(v) *= a;}
   template <typename T> inline Vector4<T> operator*(const float a, const Vector4<T> v)
      { return Vector4<T>(v) *= a;}
   template <typename T> inline Vector4<T> operator*(const double a, const Vector4<T>& v)
      { return Vector4<T>(v) *= a;}

   /////////////////////////////////////////////////////////////////////////////

   typedef Vector2<int> Vector2i;
   typedef Vector2<float> Vector2f;
   typedef Vector2<double> Vector2d;

   typedef Vector3<int> Vector3i;
   typedef Vector3<float> Vector3f;
   typedef Vector3<double> Vector3d;

   typedef Vector4<int> Vector4i;
   typedef Vector4<float> Vector4f;
   typedef Vector4<double> Vector4d;

   } // namespace arda

////////////////////////////////////////////////////////////////////////////////
template <typename T> std::string & arda::Vector2<T>::getstring(std::string & s) const
   {
   s.clear();
   std::stringstream ss;
   ss << "[ " << x << ", " << y << " ]";
   s = ss.str();
   return s;   
   }

////////////////////////////////////////////////////////////////////////////////
template <typename T> std::string & arda::Vector3<T>::getstring(std::string & s) const
   {
   s.clear();
   std::stringstream ss;
   ss << "[ " << x << ", " << y << ", " << z << " ]";
   s = ss.str();
   return s;   
   }

////////////////////////////////////////////////////////////////////////////////
template <typename T> std::string & arda::Vector4<T>::getstring(std::string & s) const
   {
   s.clear();
   std::stringstream ss;
   ss << "[ " << x << ", " << y << ", " << z << ", " << w << " ]";
   s = ss.str();
   return s;   
   }


#endif // VECTOR_H_
